/*
 * USB ICM532 WebCam driver
 * Copyright (c) 2004 Fabio Di Niro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This source code is based heavily on the vicam webcam driver which was
 * written by Joe Burks, Christopher L Cheney, Pavel Machek, John Tyner and
 * Monroe Williams
 *
 */

/* internal rev. 0.1 */
 
#include "icm532.h"
#include "icm532_camera.c"

/*
 * ---------------------------------------------------------
 *                  Camera control stuff
 * ---------------------------------------------------------
 */

/*
 * ---------------------------------------------------------
 *                  Video Device stuff
 * ---------------------------------------------------------
 */
static int
icm532_open(struct video_device *dev, int flags)
{
	/* Retrieve the icm532_camera pointer stored in the priv field */
	struct icm532_camera *cam = (struct icm532_camera *) dev->priv;
	if (!cam) {
		DBG("ICM532 video_device improperly initialized\n");
	}

	/* checks if camera is already opened */
	if (cam->is_opened) {
		DBG("ICM532 open called on already opened camera\n");
		return -EBUSY;
	}

	/* initialize camera if it has not been initalized */
	if (!cam->is_initialized) {
		DBG("ICM532 camera is not initialized, launching initialize_camera\n");
		initialize_camera(cam);
	}

	DBG("ICM532 allocate raw and framebuffer memory\n");
	if (!cam->raw_line) {
		//cam->raw_line = kmalloc(ICM532_MAX_FRAME_SIZE, GFP_KERNEL);
		cam->raw_line = usbvideo_rvmalloc(ICM532_MAX_FRAME_SIZE);
		if (!cam->raw_line) {
			printk(KERN_ERR "ICM532 could not allocate memory for raw buffer!\n");
			return -ENOMEM;
		}
	}
	/*
	if (!cam->urb_data) {
		cam->urb_data = kmalloc(ICM532_MAX_READ_SIZE, GFP_KERNEL);
		if (!cam->urb_data) {
			printk(KERN_ERR "ICM532 could not allocate memory for urb data buffer!\n");
			return -ENOMEM;
		}
	}*/
	
	if (!cam->framebuf) {
		cam->framebuf =
		    usbvideo_rvmalloc(ICM532_MAX_FRAME_SIZE * ICM532_FRAMES);
		if (!cam->framebuf) {
			printk(KERN_ERR "ICM532 could not allocate memory for frame buffer!\n");
			usbvideo_rvfree(cam->raw_line, ICM532_MAX_FRAME_SIZE);
			//kfree(cam->raw_line);
			//kfree(cam->urb_data);
			return -ENOMEM;
		}
	}
	/* 
	 * store the size of the allocated frame buffer 
	 */
	cam->framebuf_size = ICM532_MAX_FRAME_SIZE * ICM532_FRAMES;
	
	DBG("ICM532 setting is_opened to true\n");
	cam->is_opened = 1;
	return 0;
}

static void
icm532_close(struct video_device *dev)
{
	/* Retrieve the icm532_camera pointer stored in the priv field */
	struct icm532_camera *cam = (struct icm532_camera *) dev->priv;
	
	if (cam->is_capturing) stop_capture(cam);
	
	if (cam->is_removed) {
		DBG("ICM532 cam is_removed, close will launch purge\n");
		icm532_purge(cam);
	} else {
		DBG("ICM532 camera is not removed, close will only set is_opened to false\n");
		cam->is_opened = 0;
	}
}

static long
icm532_read(struct video_device *dev, char *buf,
		 unsigned long count, int noblock)
{
	/* 
	 * Retrieve the icm532_camera pointer stored in the priv field 
	 */
	struct icm532_camera *cam = (struct icm532_camera *) dev->priv;
	
	if (!buf)
		return -EINVAL;

	if (!count)
		return -EINVAL;
	DBG("ICM532 icm532_read %d bytes.\n", (int) count);
	read_frame(cam, 0);
	// FABBB
	// This is some code that will hopefully allow us to do shell copies from
	// the /dev/videoX to a file and have it actually work.
	if (cam->framebuf_size != 0) {
		if (count <= cam->framebuf_size) {
			// count does not exceed available bytes
			if (copy_to_user(buf, (cam->framebuf), count))
				return -EFAULT;
			return count;
		} else {
			count = cam->framebuf_size;
			if (copy_to_user(buf, (cam->framebuf), count))
				return -EFAULT;
			return count;
		}
	}
	
	

	if (count > cam->framebuf_size)
		count = cam->framebuf_size;

	if (copy_to_user(buf, cam->framebuf, count)) {
		return -EFAULT;
	}
	return count;
}

static long
icm532_write(struct video_device *v, const char *buf, unsigned long count,
		  int nonblock)
{
	DBG("ICM532 function icm532_write\n");

	return -EINVAL;
}

static int
icm532_ioctl(struct video_device *dev, unsigned int ioctlnr, void *arg)
{
	/* Retrieve the icm532_camera pointer stored in the priv field */
	struct icm532_camera *cam = (struct icm532_camera *) dev->priv;
	
	int retval = 0;

	if (!cam)
		return -ENODEV;
	

	switch (ioctlnr) {
		
	/* video device query capabilities */
	case VIDIOCGCAP:	// FABBB
		{
			struct video_capability b;

			DBG("ICM532 VIDIOCGCAP\n");
			strcpy(b.name, "ICM532 based Camera");
			b.type = VID_TYPE_CAPTURE;
			b.channels = 1;
			b.audios = 0;
			b.maxwidth = 352;	/* VIDEOSIZE_CIF */
			b.maxheight = 288;
			b.minwidth = 352;	/* VIDEOSIZE_48_48 */
			b.minheight = 288;

			if (copy_to_user(arg, &b, sizeof (b)))
				retval = -EFAULT;

			break;
		}

	/* get video source channel - none available */
	case VIDIOCGCHAN:
		{
			struct video_channel v;

			DBG("ICM532 VIDIOCGCHAN\n");
			if (copy_from_user(&v, arg, sizeof (v))) {
				retval = -EFAULT;
				break;
			}
			if (v.channel != 0) {
				retval = -EINVAL;
				break;
			}

			v.channel = 0;
			strcpy(v.name, "Camera");
			v.tuners = 0;
			v.flags = 0;
			v.type = VIDEO_TYPE_CAMERA;
			v.norm = 0;

			if (copy_to_user(arg, &v, sizeof (v)))
				retval = -EFAULT;
			break;
		}
		
	/* set video source channel - none available */
	case VIDIOCSCHAN:
		{
			int v;

			if (copy_from_user(&v, arg, sizeof (v)))
				retval = -EFAULT;
			DBG("ICM532 VIDIOCSCHAN %d\n", v);

			if (retval == 0 && v != 0)
				retval = -EINVAL;

			break;
		}

	/* get image properties */
	case VIDIOCGPICT:
		{
			struct video_picture vp;
			DBG("ICM532 VIDIOCGPICT\n");
			memset(&vp, 0, sizeof (struct video_picture));
			// TODO vp.brightness = cam->gain << 8; FABBB
			vp.depth = 24;
			vp.palette = VIDEO_PALETTE_RGB24;
			if (copy_to_user
			    (arg, &vp, sizeof (struct video_picture)))
				retval = -EFAULT;
			break;
		}

	/* set image properties */
	case VIDIOCSPICT:
		{
			struct video_picture vp;

			if(copy_from_user(&vp, (struct video_picture *) arg,
				sizeof(struct video_picture)))
				retval = -EFAULT;
			else
			{
				DBG("ICM532 VIDIOCSPICT depth = %d, pal = %d\n", vp.depth,
				    vp.palette);

				// TODO cam->gain = vp.brightness >> 8; FABBB

				if (vp.depth != 24
				    || vp.palette != VIDEO_PALETTE_RGB24)
					retval = -EINVAL;
			}
			break;
		}

	/* get capture window */
	case VIDIOCGWIN:
		{
			struct video_window vw;
			vw.x = 0;
			vw.y = 0;
			vw.width = 352;
			vw.height = 288;
			vw.chromakey = 0;
			vw.flags = 0;
			vw.clips = NULL;
			vw.clipcount = 0;

			DBG("ICM532 VIDIOCGWIN\n");

			if (copy_to_user
			    ((void *) arg, (void *) &vw, sizeof (vw)))
				retval = -EFAULT;

			/* I'm not sure what the deal with a capture window is, it is very poorly described
			 * in the doc.  So I won't support it now.

			 * FABBB so do I :-)
			 */
			break;
		}

	/* set capture window */
	case VIDIOCSWIN:
		{

			struct video_window *vw = (struct video_window *) arg;
			DBG("ICM532 VIDIOCSWIN %d x %d\n", vw->width, vw->height);

			if ( vw->width != 352 || vw->height != 288 )
				retval = -EFAULT;
			break;
		}

	/* mmap interface TODO FABBB */
	case VIDIOCGMBUF:
		{
			struct video_mbuf vm;
			int i;

			DBG("ICM532 VIDIOCGMBUF\n");
			memset(&vm, 0, sizeof (vm));
			vm.size =
			    ICM532_MAX_FRAME_SIZE * ICM532_FRAMES;
			vm.frames = ICM532_FRAMES;
			for (i = 0; i < ICM532_FRAMES; i++)
				vm.offsets[i] = ICM532_MAX_FRAME_SIZE * i;

			if (copy_to_user
			    ((void *) arg, (void *) &vm, sizeof (vm)))
				retval = -EFAULT;
			break;
		}

	/* start capture of the frame */	
	case VIDIOCMCAPTURE:
		{
			struct video_mmap vm;
			// int video_size;

			if (copy_from_user
			    ((void *) &vm, (void *) arg, sizeof (vm))) {
				retval = -EFAULT;
				break;
			}
			DBG("ICM532 VIDIOCMCAPTURE frame#%d, height=%d, width=%d, format=%d.\n",vm.frame,vm.width,vm.height,vm.format);

			if ( vm.frame >= ICM532_FRAMES || vm.format != VIDEO_PALETTE_RGB24 )
				retval = -EINVAL;

			if (!cam->is_capturing) {
				DBG("VIDIOCMCAPTURE with cam not capturing, launch start_capture\n");
				//retval = -EFAULT;
				cam->have_to_stop = 0;
				if (start_capture(cam)) {
					DBG("Failed to start capture\n");
					retval = -EFAULT;
					break;
				}
			};
			cam->raw_requested[vm.frame] = 1;
			break;
		}

	/* wait until the requested frame is available */
	case VIDIOCSYNC:
		{
			int frame, i;
			
			if (copy_from_user((void *) &frame, arg, sizeof (int))) {
				retval = -EFAULT;
				break;
			}
			DBG("ICM532 VIDIOCSYNC: %d, cam capturing:%d cam have to stop:%d\n", frame, cam->is_capturing, cam->have_to_stop);
			
			//if (! cam->raw_requested[frame]) {
			//	DBG("VIDIOCSYNC on a frame not requested\n");
			//	retval = -EINVAL;
			//	break;
			//}
			i = wait_event_interruptible(cam->wait_queue, cam->raw_ready[frame]);
			if (i) DBG("VIDIOCSYNC did not caught frame %d ready\n", frame);
			DBG("VIDIOCSYNC got frame %d ok\n", frame);
			cam->raw_ready[frame] = 0;
			break;
		}

		/* pointless to implement overlay with this camera */
	case VIDIOCCAPTURE:
	case VIDIOCGFBUF:
	case VIDIOCSFBUF:
	case VIDIOCKEY:
		retval = -EINVAL;
		break;

		/* tuner interface - we have none */
	case VIDIOCGTUNER:
	case VIDIOCSTUNER:
	case VIDIOCGFREQ:
	case VIDIOCSFREQ:
		retval = -EINVAL;
		break;

		/* audio interface - we have none */
	case VIDIOCGAUDIO:
	case VIDIOCSAUDIO:
  		retval = -EINVAL;
		break;
	default:
  		retval = -ENOIOCTLCMD;
		break;
	}

	return retval; //should return a decent value :-S
}

static int
icm532_mmap(struct video_device *dev, const char *adr, unsigned long size)
{
	struct icm532_camera *cam = dev->priv;
	unsigned long start = (unsigned long) adr;
	unsigned long page, pos;

	DBG("ICM532 function icm532_mmap %ld\n", size);

	if (!cam) return -ENODEV;

	if (!cam->framebuf) {	/* FABBB we do lazy allocation */
		cam->framebuf =
		    usbvideo_rvmalloc(ICM532_MAX_FRAME_SIZE * ICM532_FRAMES);
		if (!cam->framebuf) {
			return -ENOMEM;
		}
	}

	pos = (unsigned long) (cam->framebuf);
	while (size > 0) {
		page = usbvideo_kvirt_to_pa(pos);

		if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED)) {
			return -EAGAIN;
		}
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}

	return 0;

}

int
icm532_video_init(struct video_device *vdev)
{
	DBG("ICM532 function icm532_video_init\n");

	// This would normally create the proc entry for this camera

	return 0;
}

static void
read_frame(struct icm532_camera *cam, int framenum)
{
	DBG("ICM532 read_frame\n");

	int i;
	cam->have_to_stop = 0;
	if (start_capture(cam)) {
		DBG("Failed to start capture\n");
		return -EFAULT;
	}
	cam->raw_requested[0] = 1;

	/* wait until the requested frame is available */
	i = wait_event_interruptible(cam->wait_queue, cam->raw_ready[0]);
	if (i) DBG("read_fram did not caught frame 0 ready\n");
	DBG("read_fram got frame 0 ok\n");
	cam->raw_ready[0] = 0;
	cam->have_to_stop = 1;



	return;
}




/*
 * -----------------------------------------------------------
 *           USB Driver stuff
 * -----------------------------------------------------------
 */
static void *
icm532_probe(struct usb_device *dev, unsigned int ifnum,
		  const struct usb_device_id *id)
{
	struct icm532_camera *cam;
	DBG("ICM532 function icm532_probe\n");

	/* See if the device offered us matches what we can accept */
	if ((dev->descriptor.idVendor != USB_ICM532_VENDOR_ID) ||
	    (dev->descriptor.idProduct != USB_ICM532_PRODUCT_ID)) {
	    	printk(KERN_WARNING "ICM532 the offered device is not an icm532\n");
		return NULL;
	}
	
	printk(KERN_WARNING "ICM532 based webcam connected\n");

	DBG(KERN_DEBUG "ICM532 starting resource allocation\n");

	/* Allocate the icm532_camera struct in kernel memory */
	if ((cam = kmalloc(sizeof (struct icm532_camera), GFP_KERNEL)) == NULL) {
		printk(KERN_WARNING "ICM532 could not allocate kernel memory for icm532_camera struct\n");
		return NULL;
	} else DBG(KERN_DEBUG "ICM532 icm532_camera struct allocated in kernel memory!\n");

	/* zerofill of the freshly allocated struct */
	memset(cam, 0, sizeof (struct icm532_camera));

	/* copy the v4l icm532_video_device struct in the vdev field of the icm532_camera struct */
	memcpy(&cam->vdev, &icm532_video_device,
	       sizeof (icm532_video_device));
	
	/* a sort of reverse mapping for those functions that get vdev only */
	cam->vdev.priv = cam;
	DBG(KERN_DEBUG "ICM532 initializing vdev and udev\n");
	/* set the usb device pointer in the icm532_camera struct */
	cam->udev = dev;
	
	DBG("Initializing wait queue\n");
	init_waitqueue_head(&cam->wait_queue);
	
	

	DBG(KERN_DEBUG "ICM532 trying to register video device\n");
	
	/* Now we are ready for register our video device */
	if (video_register_device(&cam->vdev, VFL_TYPE_GRABBER, -1) == -1) {
		kfree(cam);
		printk(KERN_WARNING "ICM532 video_register_device failed\n");
		return NULL;
	}
	
	printk(KERN_INFO "ICM532 webcam driver now controlling video device %d\n",cam->vdev.minor);
	if (!cam->raw_line) {
		//cam->raw_line = kmalloc(ICM532_MAX_FRAME_SIZE, GFP_KERNEL);
		cam->raw_line = usbvideo_rvmalloc(ICM532_MAX_FRAME_SIZE);
		if (!cam->raw_line) {
			printk(KERN_ERR "ICM532 could not allocate memory for raw buffer!\n");
			return -ENOMEM;
		}
	}
	/*if (!cam->urb_data) {
		cam->urb_data = kmalloc(ICM532_MAX_READ_SIZE, GFP_KERNEL);
		if (!cam->urb_data) {
			printk(KERN_ERR "ICM532 could not allocate memory for urb data buffer!\n");
			return -ENOMEM;
		}
	}*/
	initialize_camera(cam);
	
	return cam;
}

static void
icm532_purge(struct icm532_camera *cam)
{
	DBG("ICM532 unregistering video device %d\n", cam->vdev.minor);
	video_unregister_device(&cam->vdev);

	if (cam->raw_line)
		//kfree(cam->raw_line);
		usbvideo_rvfree(cam->raw_line,
		       ICM532_MAX_FRAME_SIZE);
	if (cam->urb_data)
		kfree(cam->urb_data);
	if (cam->framebuf)
		usbvideo_rvfree(cam->framebuf,
		       ICM532_MAX_FRAME_SIZE * ICM532_FRAMES);
	if (cam->is_capturing)
	{
		stop_capture(cam);
	}
	DBG("ICM532 freeing icm532_camera instance in kernel memory\n");
	kfree(cam);

	printk(KERN_DEBUG "ICM532 based WebCam disconnected\n");
}

static void
icm532_disconnect(struct usb_device *dev, void *ptr)
{
	struct icm532_camera *cam = ptr;
	if (cam->is_capturing) {
		DBG("ICM532 cam is capturing, launching stop_capture\n");
	}
	
	if (cam->is_opened) {
		DBG(KERN_DEBUG "ICM532 cam is opened, disconnect will only set is_removed to true\n");
		cam->is_removed = 1;
	} else {
		DBG(KERN_DEBUG "ICM532 cam is not opened, disconnect will launch purge\n");
		icm532_purge(cam);
	}
}

/*
 * -----------------------------------------------------------
 *           Kernel module stuff
 * -----------------------------------------------------------
 */
 
static int icm532_init(void)
{
	DBG("ICM532 function icm532_init\n");
	printk(KERN_WARNING "ICM532 WebCam driver start\n");
	DBG("ICM532 registering driver\n");
	if (usb_register(&icm532_driver) != 0)
		printk(KERN_WARNING "ICM532 usb_register failed!\n");
	return 0;
}


static void icm532_exit(void)
{
	DBG("ICM532 function icm532_exit\n");

	DBG("ICM532 deregistering driver\n");

	usb_deregister(&icm532_driver);
	printk(KERN_WARNING "ICM532 WebCam driver unload\n");
}
