/*
 * USB ICM532 WebCam driver
 * Copyright (c) 2004 Fabio Di Niro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * This source code is based heavily on the vicam webcam driver which was
 * written by Joe Burks, Christopher L Cheney, Pavel Machek, John Tyner and
 * Monroe Williams
 *
 */
 
/* internal rev. 0.3 */
/* This file (icm532_camera.c) is just an outtake from icm532.c
	I did this just for concentrating on the portion of the driver
	that was working less...
	As soon as the driver will show us an image, I'll reintegrate
	this bunch of function in the main source file...
 */
// -----------------------------------------------------------
//           read and writes sequences
// -----------------------------------------------------------


static int write_byte_register (struct usb_device *dev, u8 icm_register, u8 value)
{
	int status;
	/* from the IC Media specs, if lenght = 0 value contains the value to be written,
		if lenght > 0 (<=8) value is zero and the transfer buffer contains the data to
		be written into the register */
	/* the usb sniff report register writes of one byte per write always passed with a 
		transfer buffer (lenght = 1). I think that this is for writing just 1 byte into
		register (value is a word) */
	status = usb_control_msg (dev, usb_sndctrlpipe( dev, 0), 0x02, USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE, 0x00, icm_register, &value, 0x01, HZ);
	return status;
}

static int read_byte_register (struct usb_device *dev, u8 icm_register, u8 *value)
{
	int status;
	status = usb_control_msg(dev, usb_rcvctrlpipe( dev, 0 ), 0x03, USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE , 0x00, icm_register, value, 0x01, HZ);
	return status;
}

int urb_message(struct icm532_camera *cam, icm532_urb_t message)
{
	u_int8_t value;
	int i, status;
	
	switch ( message.action ) {
		case WRITE_URB:
			status = write_byte_register(cam->udev, message.icm_register, message.value);
			if (status >= 0) {
				status = 0;
			} else {
				DBG("write of register %2.0X failed\n", message.value);
			}
			return status;
		break;
		
		case READ_URB:
			read_byte_register(cam->udev, message.icm_register, &value);
			if (value != message.value) DBG("read of register %2.0X got %2.0X instead of %2.0X\n", message.icm_register, value, message.value);
			return 0;
		break;
		
		case WAIT_VALUE:
			for (i = 0; i < 100; i++) {
				read_byte_register(cam->udev, message.icm_register, &value);
				if (value == message.value) {
					DBG("WAIT_VALUE: got value %2.0X in register %2.0X, OK\n", value, message.icm_register);
					return 0;
				}
			}
			return -1;
		break;
		
		case SET_INTERFACE:
			status = usb_set_interface (cam->udev, 0, message.value);
			if (status) DBG("set alternate %d failed, status %d\n", message.value, status);
			return status;
		break;
		
		case RESET_PIPE:
			status = usb_clear_halt(cam->udev, message.value);
			if (status) DBG("reset pipe failed, status %d\n", status);
			return status;
		break;
		default:
				return -1;
		break;
	}
}


static int
initialize_camera(struct icm532_camera *cam)
{
	int status, i;

	DBG("ICM532 initializing camera\n");
	DBG("ICM532 resetting device\n");
	status = usb_reset_device(cam->udev);
	if (status < 0) {
		DBG("ICM532 FAILED reset device, status %d\n", status);
	};

	for(i = 0; i < INIT_URBS; i++) {
		status = urb_message(cam, icm532_init_sequence[i]);
		if (status != 0) DBG("error in control urb #%d of init sequence\n", i);
	}
	cam->is_initialized = 1;
	return status;
};



static int start_capture(struct icm532_camera *cam)
{
	int i, j, status;
	int pacchetti = 32;
	int dim_packet = 1023;
	int dim_buffer = dim_packet * pacchetti;
	struct urb *urb;
	int fx, err, frame_now;
	
	
	DBG("ICM532 start_capture\n");
	cam->number=0;
	
	for(i = 0; i < START_CAPTURE_URBS; i++) {
		status = urb_message(cam, icm532_start_capture_sequence[i]);
		if (status != 0) DBG("error in control urb #%d of start capture sequence\n", i);
	}

	cam->is_capturing = 1;
	// allocate the urb data buffer
	if (!cam->urb_data) {
		cam->urb_data = kmalloc(ICM532_MAX_READ_SIZE, GFP_KERNEL);
		if (!cam->urb_data) {
			printk(KERN_ERR "ICM532 could not allocate memory for urb data buffer!\n");
			return -ENOMEM;
		}
	}
	
	// allocate the start needed urbs 
	if (!cam->start_urb[0]) {
		for (i=0; i < 2; i++) {
			DBG("ICM532 allocating start urb #%d\n", i);
			urb = usb_alloc_urb(32);
			if (!urb) {
				DBG("ICM532 usb_alloc_urb returned an error NULL\n");
				return -ENOMEM;
			}
			cam->start_urb[i] = urb;
		}
	} 
	
	// allocate all the needed urbs (10)
	if (!cam->iso_urb[0]) {
		for (i=0; i < 10; i++) {
			DBG("ICM532 allocating urb #%d\n", i);
			urb = usb_alloc_urb(32);
			if (!urb) {
				DBG("ICM532 usb_alloc_urb returned an error NULL\n");
				return -ENOMEM;
			}
			cam->iso_urb[i] = urb;
		}
	} 

	// prepare the 10 urbs for isoch transfer
	for (i=0; i < 10; i++) {
		urb = cam->iso_urb[i];
		urb->dev = cam->udev;
		urb->context = cam;
		urb->pipe = usb_rcvisocpipe(cam->udev, 0x81);	
		urb->transfer_buffer = cam->urb_data;
		urb->transfer_flags = USB_ISO_ASAP;
		urb->complete = urb_iso_complete;
		urb->number_of_packets = 32;
		urb->transfer_buffer_length = dim_buffer;
		urb->interval = 1;
		for (fx = 0; fx < pacchetti; fx++) {
			urb->iso_frame_desc[fx].offset = dim_packet * fx;
			urb->iso_frame_desc[fx].length = dim_packet;
		}
	}	
	
	// get the current frame number
	frame_now = usb_get_current_frame_number(cam->udev);

	// prepare the start urbs
	for (i=0; i < 2; i++) {
		urb = cam->start_urb[i];
		urb->dev = cam->udev;
		urb->context = cam;
		urb->pipe = usb_rcvisocpipe(cam->udev, 0x81);	
		urb->transfer_buffer = cam->urb_data;
		urb->complete = urb_iso_complete;
		urb->number_of_packets = 32;
		urb->transfer_buffer_length = dim_buffer;
		urb->interval = 1;
		for (fx = 0; fx < pacchetti; fx++) {
			urb->iso_frame_desc[fx].offset = dim_packet * fx;
			urb->iso_frame_desc[fx].length = dim_packet;
		}
	}
	// set the start time of the first two urbs
	cam->start_urb[0]->start_frame = frame_now + 10;
	cam->start_urb[1]->start_frame = frame_now + 42;
	DBG("ISOC actual frame time: %d\n", frame_now);
	DBG("ISOC #0 will start at %d\n", cam->start_urb[0]->start_frame);
	DBG("ISOC #1 will start at %d\n", cam->start_urb[1]->start_frame);
	for (i = 0; i < 2; i++) {
		err = usb_submit_urb(cam->start_urb[i]);
		if (err) {
			DBG("usb_submit_urb #%d returned error %d\n", i, err);
			for (j = i-1; j >= 0; j--)
				if (!usb_unlink_urb(cam->start_urb[j]))
					usb_free_urb(cam->start_urb[j]);
			return err;
		}
		
		cam->number += 1;
		cam->urbs_onair += 1;
		DBG("sent urb #%d, %d onair\n", cam->number, cam->urbs_onair);
	}
	cam->is_capturing = 1;
	cam->urb_sequence = 0;
	cam->raw_in_progress = 0;
	cam->raw_requested[0] = 0;
	cam->raw_requested[1] = 0;
	cam->raw_ready[0] = 0;
	cam->raw_ready[1] = 0;
	urb_iso_sequence(cam);
	return 0;
};

static int stop_capture(struct icm532_camera *cam)
{
	int i;
	
	DBG("stop_capture\n");
	if (! cam->is_capturing) {
		DBG("camera was not capturing!\n");
		return 0;
	}
	cam->have_to_stop = 1;
	i = wait_event_interruptible(cam->wait_queue, cam->urbs_onair == 0);
	if (i) DBG("there are still urbs onair\n");
	DBG("exited from wait queue, urbs onair: %d\n", cam->urbs_onair);
	
	for (i = 0; i < 10; i++) 
		if (cam->iso_urb[i]) 
		{
			usb_free_urb(cam->iso_urb[i]);
			DBG("freeing urb #%d\n", i);
		}
		
	for (i = 0; i < 2; i++) 
		if (cam->start_urb[i]) 
		{
			usb_free_urb(cam->start_urb[i]);
			DBG("freeing start urb #%d\n", i);
		}
		
	if (cam->urb_data) {
		DBG("freeing urb data buffer\n");
		kfree(cam->urb_data);
	}
	
	cam->have_to_stop = 0;
	cam->is_capturing = 0;
	return 0;
};


static int urb_iso_sequence(struct icm532_camera *cam)
{
	int pacchetti = 32;
	struct urb *urb;
	int err, i, j;

	if (!cam->is_capturing) {
		DBG("ICM532 request iso when camera was not capturing, launching start_capture\n");
		start_capture(cam);
	}
	
	for (i = 5 * cam->urb_sequence; i < (5 * cam->urb_sequence + 5); i++) {
		urb = cam->iso_urb[i];
		urb->dev = cam->udev;
		/* Resets the urb state */
		urb->status = 0;
		for (j = 0; j < pacchetti; j++) urb->iso_frame_desc[j].status = 0;
		/* and resubmit it */
		err = usb_submit_urb(cam->iso_urb[i]);
		if (err) {
			DBG("usb_submit_urb #%d returned error %d\n", i, err);
			for (j = i-1; j >= 5*cam->urb_sequence; j--)
				if (!usb_unlink_urb(cam->iso_urb[j]))
					usb_free_urb(cam->iso_urb[j]);
			return err;
		}
		cam->number += 1;
		cam->urbs_onair += 1;
	}
	
	/* Sets next urb sequence to submit */
	if (cam->urb_sequence) {
		cam->urb_sequence = 0;
	} else {
		cam->urb_sequence = 1;
	}
	return 0;
}



static void urb_iso_complete(struct urb *urb)
{
	struct icm532_camera *cam;
	int i, bytes_total = 0;

	if (!urb->context) {
		DBG("ICM532 no context\n");
		return;
	}
	cam = urb->context;
	cam->urbs_onair -=1;
	
	// request another urb sequence if running low...
	if ((cam->urbs_onair < 2) && (!cam->have_to_stop)) {
		urb_iso_sequence(cam);
	} 
	
	/* Wake up stop_capture if invoked */
	if ((cam->urbs_onair == 0) && (cam->have_to_stop) && (waitqueue_active(&cam->wait_queue)))
		wake_up_interruptible(&cam->wait_queue);
	
	switch (urb->status) {
		case 0:
			//DBG("ISOC #%d status OK\n", cam->number);
		break;
		case -EINPROGRESS:
			DBG("status EINPROGRESS\n");
			return;
		break;
		case -ECONNRESET:		// unlink
			DBG("status ECONNRESET\n");
			return;
		break;
		case -ESHUTDOWN:		// disconnect etc
			DBG ("status ESHUTDOWN\n");
			return;
		break;
		case -ECONNABORTED:		// endpoint reset
			DBG("status ECONNABORTED\n");
			return;
		break;
		case -EOVERFLOW:		// data overrun 
			DBG("status EOVERFLOW\n");
			return;
		break;
		default:
			DBG("status unknown %d\n", urb->status);
			return;
		break;
	}
	
	for (i=0; i < urb->number_of_packets; i++) {

		if (urb->iso_frame_desc[i].actual_length > 4) {
			/* Frame start */
			if ((*(cam->urb_data + urb->iso_frame_desc[i].offset) == 0x80)) {
				DBG("new frame start, lines %d\n", cam->decoded_lines);
				stop_grab(cam);
				cam->sequence = 0x80;
				cam->decoded_lines = 0;
				/* if frame was requested, clear its ready flag and start grabbing */
				if (cam->raw_requested[cam->raw_in_progress]) {
					DBG("frame %d was requested, clearing ready state, start grabbing\n", cam->raw_in_progress);
					cam->raw_ready[cam->raw_in_progress] = 0;
				}
			}
			if (cam->decoded_lines > 64) stop_grab(cam);
				
			memcpy(cam->raw_line + (cam->decoded_lines * 352 * 3), cam->urb_data + urb->iso_frame_desc[i].offset, urb->iso_frame_desc[i].actual_length);
			cam->decoded_lines +=1;
			
			bytes_total += urb->iso_frame_desc[i].actual_length;
			cam->sequence +=1;
		}
	}
	
	cam->raw_size[cam->raw_in_progress] += bytes_total;
	return;
};


/*  
 * bayer to RGB conversion function derived from the one by Takafumi Mizuno  
 * A raw line coming from the cam is about 3 and half bayer lines.
 * This function process the 3 whole bayer lines which are supposed to be
 * 1st line RGRG
 * 2nd line GBGB
 * 3rd line RGRG

 * Bayer pattern 
 * RG RG RG RG
 * GB GB GB GB
 *
 * RG RG RG RG
 * GB GB GB GB
 */
void new_bayer2rgb(unsigned char *dst, unsigned char *src) {
	unsigned char *imgpt;
	int i, gap, scan_line;
	int interval = 323;
	imgpt = dst;
//#define RAWOUT
#ifdef RAWOUT
	memcpy(dst, src, 1023);
	return;
#endif	
	for (i = 2; i < (interval + 2); i++) { /* First line is a RGRG */
		if ((i % 2) == 0) {
			/* (R)G
			 *  G B
			 */
			*imgpt++ = *(src + i);									/* R */
			*imgpt++ = (*(src + i + 1) + *(src + i + interval))/2;		/* G */ 
			*imgpt++ = *(src + i + interval - 1) + *(src + i +interval + 1);	/* B */ 
		} else {
			/*  R(G)
			 *  G B
			 */
			*imgpt++ = (*(src + i - 1) + *(src + i + 1)) /2;		/* R */
			*imgpt++ = (*(src + i) + *(src + i + interval)) / 2;			/* G */ 
			*imgpt++ = *(src + i + interval);							/* B */ 
		}
	}
	gap = imgpt - dst;
	scan_line = gap / (352 * 3);
	scan_line ++;
	gap = scan_line * 352 * 3;
	imgpt = dst + gap;
	for (i = (interval + 2); i < (interval + interval + 2); i++) {	/* Second line is a GBGB */
		if ((i % 2) == 0) {
			/*  R G
			 * (G)B
			 */
			*imgpt++ = (*(src + i - interval) + *(src + i +interval)) / 2;	/* R */
			*imgpt++ = (*(src + i) + *(src + i - interval + 1))/2;		/* G */
			*imgpt++ = (*(src + i + 1) + *(src + i - 1)) / 2;		/* B */
		} else {
			/*  R G
			 *  G(B)
			 */
			*imgpt++ = (*(src + i - interval - 1) + *(src + i - interval + 1)) / 2;/* R */
			*imgpt++ = (*(src + i - 1) + *(src + i + 1)) / 2;			/* G */
			*imgpt++ = *(src + i);/* B */
		
		}
	}
	gap = imgpt - dst;
	scan_line = gap / (352 * 3);
	scan_line ++;
	gap = scan_line * 352 * 3;
	imgpt = dst + gap;
	for (i = (interval + interval + 2); i < (interval + interval + interval + 2); i++) {	/* Third line is a RGRG */
		if ((i % 2) == 0) {
			/* (R)G
			 *  G B
			 */
			*imgpt++ = *(src + i);									/* R */
			*imgpt++ = (*(src + i + 1) + *(src + i - 1))/2;		/* G */ 
			*imgpt++ = *(src + i - interval - 1) + *(src + i - interval + 1);	/* B */ 
		} else {
			/*  R(G)
			 *  G B
			 */
			*imgpt++ = (*(src + i - 1) + *(src + i + 1)) /2;		/* R */
			*imgpt++ = (*(src + i) + *(src + i - interval - 1)) / 2;		/* G */ 
			*imgpt++ = *(src + i - interval);							/* B */ 
		}
	}

	return;	
}


void dump_buffer(u8 *buffer, int lenght) {
	int bytes_per_line = 32;
	int i;
	for(i = 0; i < lenght; i++) {
		printk("%2.0X ", *(buffer + i));
		if (!(i % bytes_per_line) && i) printk("\n");
	};
	printk("\n");
};

void stop_grab(struct icm532_camera *cam){
	int lines;
	DBG("frame %d ended with %d bytes, %d lines decoded, last sequence %2.0x\n", cam->raw_in_progress, cam->raw_size[cam->raw_in_progress], cam->decoded_lines, cam->sequence);
	/* Process image complete, frame cam->raw_in_progress READY!!! */
	
	if (cam->decoded_lines < 30) {
		/* frame is not good */
		cam->sequence = 0x80;
		cam->decoded_lines = 0;
		return;
	}
	/* 
	 * if frame was requested, then mark it as ready, wake up VIDIOCSYNC and clear
	 * requested flag 
	 */
	if (cam->raw_requested[cam->raw_in_progress]) {
		for (lines = 0; lines < 64; lines ++) {
			new_bayer2rgb(cam->framebuf + (cam->raw_in_progress * ICM532_MAX_FRAME_SIZE) + (lines * 352 * 3 * 3), cam->raw_line + (lines * 352 * 3));
		}
		cam->raw_ready[cam->raw_in_progress] = 1;
		// Wake up VIDIOCSYNC if active
		if (waitqueue_active(&cam->wait_queue)) wake_up_interruptible(&cam->wait_queue);
		cam->raw_requested[cam->raw_in_progress] = 0;
	}
	
	/* set the new raw in progress */
	cam->raw_in_progress = (!cam->raw_in_progress);
	cam->raw_size[cam->raw_in_progress] = 0;
}
